import re
from functools import lru_cache
from typing import Dict, FrozenSet, List, Match, Sequence, Tuple

SUBMAP_TYPE = Sequence[Tuple[str, str]]
"""Substitution mapping type."""

COLOR_MAP: Dict[str, str] = {
    "b": "\033[94m",
    "c": "\033[96m",
    "g": "\033[92m",
    "m": "\033[95m",
    "r": "\033[91m",
    "y": "\033[93m",
    "w": "\033[97m",
    "db": "\033[34m",
    "dc": "\033[36m",
    "dg": "\033[32m",
    "dm": "\033[35m",
    "dr": "\033[31m",
    "dy": "\033[33m",
}

COLOR_END: str = "\033[0m"
# Assemble list of all color sequences
ALL_SEQ: FrozenSet[str] = frozenset([x for x in COLOR_MAP.values()] + [COLOR_END])


class MultiReplace:
    """
    MultiReplace is a tool for doing multiple find/replace actions in one pass.

    Given a mapping of values to be replaced it allows for all of the matching values to be replaced in a single pass
    which can save a lot of performance on very large strings. In addition to simple replace, it also allows for
    replacing based on regular expressions.

    Keyword Arguments:

    :type regex: bool
    :param regex: Treat search keys as regular expressions [Default: False]
    :type flags: int
    :param flags: flags to pass to the regex engine during compile

    Usage::

        from boltons import stringutils
        s = stringutils.MultiReplace((
            ('foo', 'zoo'),
            ('cat', 'hat'),
            ('bat', 'kraken)'
        ))
        new = s.sub('The foo bar cat ate a bat')
        new == 'The zoo bar hat ate a kraken'
    """

    def __init__(self, sub_map: SUBMAP_TYPE) -> None:
        """Compile any regular expressions that have been passed."""
        self.group_map: Dict[str, str] = {}
        regex_values: List[str] = []

        for idx, vals in enumerate(sub_map):
            group_name = f"group{idx}"
            exp = re.escape(vals[0])
            regex_values.append(f"(?P<{group_name}>{exp})")
            self.group_map[group_name] = vals[1]

        self.combined_pattern = re.compile("|".join(regex_values))

    def _get_value(self, match: Match) -> str:
        """Given a match object find replacement value."""
        group_dict = match.groupdict()
        key = [x for x in group_dict if group_dict[x]][0]
        return self.group_map[key]

    def sub(self, text: str) -> str:
        """Run substitutions on the input text."""
        try:
            return self.combined_pattern.sub(self._get_value, text)
        except IndexError:
            # There are no matches so just return the original text
            return text


def multi_replace(text: str, sub_map: SUBMAP_TYPE) -> str:
    """
    Shortcut function to invoke MultiReplace in a single call.

    Example Usage::

        from boltons.stringutils import multi_replace
        new = multi_replace(
            'The foo bar cat ate a bat',
            {'foo': 'zoo', 'cat': 'hat', 'bat': 'kraken'}
        )
        new == 'The zoo bar hat ate a kraken'
    """
    m = MultiReplace(sub_map)
    return m.sub(text)


@lru_cache(maxsize=None)
def get_strip_map() -> Tuple[Tuple[str, str], ...]:
    return tuple((x, "") for x in ALL_SEQ)


def strip_color(value: str) -> str:
    """Strip all color values from a given string"""
    return multi_replace(value, get_strip_map())



from unittest import TestCase

from log_color import regex


class TestRegexes(TestCase):
    def test_color_re(self):
        """Test matching basic values"""
        good_vals = (
            "#m<magenta>",
            "#b<blue>",
            "#g<green>",
            "#y<yellow>",
            "#c<cyan>",
            "#w<white>",
            "#db<darkblue>",
            "#dc<darkcyan>",
            "#dg<darkgreen>",
            "#dm<darkmagenta>",
            "#dr<darkred>",
            "#dy<darkyellow>",
        )
        bad_vals = ("Some text", "#f<fuscia>", "#m[magenta]")
        for good_val in good_vals:
            self.assertIsNotNone(regex.COLOR_EXP.match(good_val))

        for bad_val in bad_vals:
            self.assertIsNone(regex.COLOR_EXP.match(bad_val))

    def test_color_extract(self):
        """Make sure regular expression catches all instances on findall"""
        text = (
            "This text has some very colourful words in it like: #m<magenta>, "
            " #b<blue>,  #g<green>,  #y<yellow>,  #c<cyan>, and #w<wh\nite>"
        )
        vals = regex.COLOR_EXP.findall(text)

        for item in ("#m<magenta>", "#b<blue>", "#g<green>", "#y<yellow>", "#c<cyan>", "#w<wh\nite>"):
            if item not in vals:
                raise AssertionError(f"{item} was not detected")



from unittest import TestCase

from log_color.colors import ColorStr
from log_color.lib import COLOR_MAP, strip_color


class TestColorModule(TestCase):
    def test_make_str_magenta(self):
        """Verify can color strings magenta"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["m"], force_seq=True), "\033[95mcolored\033[0m")

    def test_make_str_blue(self):
        """Verify can color strings blue"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["b"], force_seq=True), "\033[94mcolored\033[0m")

    def test_make_str_green(self):
        """Verify can color strings green"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["g"], force_seq=True), "\033[92mcolored\033[0m")

    def test_make_str_yellow(self):
        """Verify can color strings yellow"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["y"], force_seq=True), "\033[93mcolored\033[0m")

    def test_make_str_red(self):
        """Verify can color strings red"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["r"], force_seq=True), "\033[91mcolored\033[0m")

    def test_make_str_cyan(self):
        """Verify can color strings cyan"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["c"], force_seq=True), "\033[96mcolored\033[0m")

    def test_make_str_white(self):
        """Verify can color strings white"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["w"], force_seq=True), "\033[97mcolored\033[0m")

    def test_make_str_dark_magenta(self):
        """Verify can color strings dark magenta"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["dm"], force_seq=True), "\033[35mcolored\033[0m")

    def test_make_str_dark_blue(self):
        """Verify can color strings dark blue"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["db"], force_seq=True), "\033[34mcolored\033[0m")

    def test_make_str_dark_green(self):
        """Verify can color strings dark green"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["dg"], force_seq=True), "\033[32mcolored\033[0m")

    def test_make_str_dark_yellow(self):
        """Verify can color strings dark yellow"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["dy"], force_seq=True), "\033[33mcolored\033[0m")

    def test_make_str_dark_red(self):
        """Verify can color strings dark red"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["dr"], force_seq=True), "\033[31mcolored\033[0m")

    def test_make_str_dark_cyan(self):
        """Verify can color strings dark cyan"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["dc"], force_seq=True), "\033[36mcolored\033[0m")

    def test_make_str_magenta_unsupported(self):
        """Verify no color sequences on magenta"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["m"], force_seq=False), "colored")

    def test_make_str_blue_unsupported(self):
        """Verify no color sequences on blue"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["b"], force_seq=False), "colored")

    def test_make_str_green_unsupported(self):
        """Verify no color sequences on green"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["g"], force_seq=False), "colored")

    def test_make_str_yellow_unsupported(self):
        """Verify no color sequences on yellow"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["y"], force_seq=False), "colored")

    def test_make_str_red_unsupported(self):
        """Verify no color sequences on red"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["r"], force_seq=False), "colored")

    def test_make_str_cyan_unsupported(self):
        """Verify no color sequences on cyan"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["c"], force_seq=False), "colored")

    def test_make_str_white_unsupported(self):
        """Verify no color sequences on white"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["w"], force_seq=False), "colored")

    def test_make_str_dark_magenta_unsupported(self):
        """Verify no color sequences on dark magenta"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["dm"], force_seq=False), "colored")

    def test_make_str_dark_blue_unsupported(self):
        """Verify no color sequences on dark blue"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["db"], force_seq=False), "colored")

    def test_make_str_dark_green_unsupported(self):
        """Verify no color sequences on dark green"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["dg"], force_seq=False), "colored")

    def test_make_str_dark_yellow_unsupported(self):
        """Verify no color sequences on dark yellow"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["dy"], force_seq=False), "colored")

    def test_make_str_dark_red_unsupported(self):
        """Verify no color sequences on dark red"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["dr"], force_seq=False), "colored")

    def test_make_str_dark_cyan_unsupported(self):
        """Verify no color sequences on dark cyan"""
        self.assertEqual(ColorStr("colored", COLOR_MAP["dc"], force_seq=False), "colored")

    def test_plat_det(self):
        """Attempt to run color sequence support detection"""
        # This test doesn't really DO much, but it at least attempts to run the
        # code which is very slightly better than no test at all.
        self.assertIn(
            ColorStr.color_supported(),
            (True, False),
        )

    def test_strip_colors(self):
        """Make sure colors can be stripped from a string"""
        self.assertEqual(
            strip_color("\033[97mwhite\033[0m \033[96mcyan\033[0m " "\033[92mgreen\033[0m"), "white cyan green"
        )



import logging
from io import StringIO
from unittest import TestCase

from log_color.colors import ColorStr
from log_color.formatters import ColorFormatter, ColorStripper
from log_color.lib import COLOR_MAP


class TestColorFormatter(TestCase):
    def setUp(self):
        self.stream = StringIO()
        self.handler = logging.StreamHandler(self.stream)
        formatter = ColorFormatter("%(levelname)s: %(message)s")
        self.handler.setFormatter(formatter)

        self.log = logging.getLogger("test_logger")
        self.log.setLevel(logging.INFO)
        for handler in self.log.handlers:
            self.log.removeHandler(handler)
        self.log.addHandler(self.handler)

    def test_format_matches(self):
        """Test format includes expected output on preformatted strings."""
        base_str = "blue"
        blue_str = ColorStr(base_str, COLOR_MAP["b"])
        expected = f"INFO: this is {blue_str}\n"
        self.log.info(f"this is #b<{base_str}>")
        self.assertEqual(self.stream.getvalue(), expected)

    def test_format_matches_arg_manage(self):
        """Test format includes expected output when formatted with logger args."""
        base_str = "blue"
        blue_str = ColorStr(base_str, COLOR_MAP["b"])
        expected = f"INFO: this is {blue_str}\n"
        self.log.info("this is #b<%s>", base_str)
        self.assertEqual(self.stream.getvalue(), expected)

    def test_format_object(self):
        """Test formatting a non-string object."""
        expected = "INFO: {}\n"
        self.log.info({})
        self.assertEqual(self.stream.getvalue(), expected)


class TestColorStripper(TestCase):
    def setUp(self):
        self.stream = StringIO()
        self.handler = logging.StreamHandler(self.stream)
        formatter = ColorStripper("%(levelname)s: %(message)s")
        self.handler.setFormatter(formatter)

        self.log = logging.getLogger("test_logger")
        self.log.setLevel(logging.INFO)
        for handler in self.log.handlers:
            self.log.removeHandler(handler)
        self.log.addHandler(self.handler)

    def test_format_matches(self):
        """Test format includes expected output."""
        base_str = "blue"
        expected = f"INFO: this is {base_str}\n"
        self.log.info("\033[94mthis\033[0m is #b<%s>", base_str)
        self.assertEqual(self.stream.getvalue(), expected)



# Standard
from unittest import TestCase


class TestImports(TestCase):
    def test_import_color_stripper(self):
        """Test importing ColorStripper."""
        try:
            from log_color import ColorStripper  # noqa
        except ImportError:
            raise AssertionError("Unable to import ColorStripper from module")

    def test_import_color_formatter(self):
        """Test importing ColorFormatter."""
        try:
            from log_color import ColorFormatter  # noqa
        except ImportError:
            raise AssertionError("Unable to import ColorFormatter from module")



import logging
from logging import LogRecord

from .colors import ColorStr
from .lib import COLOR_MAP, multi_replace, strip_color
from .regex import COLOR_EXP


class ColorStripper(logging.Formatter):
    """
    Strip references to colored output from message.

    This formatter removes all color code text from a message. For example
    it would convert ``this #g<is> text`` to ``this is text``.
    """

    def format(self, record: LogRecord) -> str:
        """Format the message."""
        msg = super().format(record)
        msg = strip_color(msg)
        for val in COLOR_EXP.findall(msg):
            if val.startswith("#d"):
                msg = msg.replace(val, val[4:-1])
            else:
                msg = msg.replace(val, val[3:-1])
        return msg


class ColorFormatter(logging.Formatter):
    """
    Insert color sequences for logging messages.

    Convert all portions of a message that contain color codes into a ColorStr instance that matches the color map data
    then re-insert it into the message. On supported platforms this will insert a color codes around the specified
    text.

    Color sequences should be in the following format: #y<message> which will be formatted as yellow. The colors
    allowed are:

    - m (magenta)
    - b (blue)
    - c (cyan)
    - g (green)
    - y (yellow)
    - r (red)
    - w (white)

    - dm (dark magenta)
    - db (dark blue)
    - dc (dark cyan)
    - dg (dark green)
    - dy (dark yellow)
    - dr (dark red)
    """

    def format(self, record: LogRecord) -> str:
        """Format the message."""
        msg = super().format(record)
        replace_map = []
        for val in COLOR_EXP.findall(msg):
            if val.startswith("#d"):
                replace_map.append((val, ColorStr(val[4:-1], COLOR_MAP[val[1:3]])))
            else:
                replace_map.append((val, ColorStr(val[3:-1], COLOR_MAP[val[1]])))
        msg = multi_replace(msg, replace_map)
        return msg



import os
from setuptools import setup
from mypyc.build import mypycify

DO_COMPILE = os.environ.get("PYTHON_MYPY_COMPILE", None)

# NOTE: Place any modules you wish compiled here
if DO_COMPILE is not None:
    setup(
        ext_modules=mypycify([
            "src/log_color/formatters.py",
            "src/log_color/lib.py",
        ])
    )
else:
    setup()



import os
import platform
import sys
from functools import lru_cache
from typing import Optional

from .lib import COLOR_END


class ColorStr(str):
    """Subclasses string to optionally include ascii color"""

    def __new__(cls, value: str, color: str, force_seq: Optional[bool] = None) -> "ColorStr":
        if cls.color_supported(force_seq=force_seq):
            return str.__new__(cls, f"{color}{value}{COLOR_END}")
        return str.__new__(cls, value)

    @staticmethod
    @lru_cache(maxsize=None)
    def color_supported(force_seq: Optional[bool] = None) -> bool:
        """Shoddy detection of color support."""
        # If True or False, override autodetection and return.
        if force_seq is False or force_seq is True:
            return force_seq

        # Honor NO_COLOR environment variable:
        if os.environ.get("NO_COLOR", None):
            return False

        # Attempt simple autodetection
        if (hasattr(sys.stderr, "isatty") and sys.stderr.isatty()) or (
            "TERM" in os.environ and os.environ["TERM"] == "ANSI"
        ):
            if "windows" in platform.system().lower() and not ("TERM" in os.environ and os.environ["TERM"] == "ANSI"):
                return False  # Windows console, no ANSI support
            else:
                return True  # ANSI output allowed
        else:
            return False  # ANSI output not allowed
